import numpy as np
from copy import deepcopy as dpc


def comparison(value, base_value, name_value):
        next_val = value.pop(0)
        next_base_value = base_value.pop(0)
        next_name_value = name_value.pop(0)

        if len(value) == 0:
            if next_val == next_base_value:
                return [], [], [next_name_value]
            else:
                return [next_val], [next_name_value], []

        nonbase_val, nonbase_name, base_name = comparison(value, base_value, name_value)
        if next_val == next_base_value:
            base_name.append(next_name_value)
        else:
            nonbase_val.append(next_val)
            nonbase_name.append(next_name_value)
        return nonbase_val, nonbase_name, base_name


class _MolBase:
    Atoms = {"":-1, "C": 0, "H": 1, "O": 2, "N": 3, "S": 4, "Na": 5}
    AtomsMass = {"C": 12, "H": 1, "O": 16, "N": 14, "S": 32, "Na": 23}
    Atomskeys = np.array(list(Atoms.keys()))
    Atomsvals = np.array(list(Atoms.values()))

    def __default_attr(self, attr_name, size=0, inplace=False):
        attr_set = ['numb', 'pos', 'link', 'atoms', 'text', 'linktype']
        if attr_name not in attr_set:
            raise RuntimeError("Wrong attribute name: {}".format(attr_name))

        if attr_name == 'numb':
            if size == 0:
                temp = None
            else:
                temp = int(size)
            if inplace:
                self.numb = temp
            return temp

        if attr_name == 'pos':
            temp = np.zeros((3, int(size)), dtype=float)
            if inplace:
                self.pos = temp
            return temp

        if attr_name == 'link':
            temp = np.zeros((int(size), 2), dtype=int)
            if inplace:
                self.link = temp
            return temp

        if attr_name == 'linktype':
            temp = np.array(["1"]*int(size), dtype=np.dtype("<U5"))
            if inplace:
                self.linktype = temp
            return temp

        if attr_name == 'atoms':
            temp = np.array(["C"]*int(size), dtype=np.dtype("<U2"))
            if inplace:
                self.atoms = temp
            return temp

        if attr_name == 'text':
            temp = np.zeros(int(size), dtype=np.dtype((str,20)))
            if inplace:
                self.text = temp
            return temp

    @staticmethod
    def __comparison(value, base_value, name_value):
        return comparison(value, base_value, name_value)

    def __init__(self, numb=None, pos=None, link=None, atoms=None, text=None, linktype=None):
        self.numb = self.__default_attr('numb')

        self.pos = self.__default_attr('pos')

        self.link = self.__default_attr('link')

        self.linktype = self.__default_attr('linktype')

        self.atoms = self.__default_attr('atoms')

        self.text = self.__default_attr('text')


        if numb is not None:
            self._init_numb(numb)

        if pos is not None:
            self._init_pos(pos)

        if linktype is not None:
            if link is None:
                msg = "linktype shoutln't be used without link. Linktype is ignored"
                raise RuntimeWarning(msg)

        if link is not None:
            self._init_link(link, linktype)

        if atoms is not None:
            self._init_atoms(atoms)

        if text is not None:
            self._init_text(text)

        self._self_consistency()

    def _init_numb(self, numb):
        try:
            self.numb = int(numb)
        except TypeError:
            raise TypeError("numb should be int-convertable")

        self.numb = numb

        if (self.pos == self.__default_attr('pos')).all():
            self.pos = self.__default_attr('pos', numb)

        if (self.atoms == self.__default_attr('atoms')).all():
            self.atoms = self.__default_attr('atoms', numb)

        if (self.text == self.__default_attr('text')).all():
            self.text = self.__default_attr('text', numb)

    def _init_pos(self, pos):
        temp = np.array(pos)
        try:
            temp = temp.astype(float)
        except TypeError:
            raise TypeError("Can't convert pos types to Mol")

        if len(temp.shape) > 2:
            raise TypeError("Wrong dimensity of pos")

        if temp.shape[0] != 3:
            if temp.shape[1] == 3:
                temp = np.transpose(temp)
            else:
                raise TypeError("Wrong shape of pos")

        if self.numb != self.__default_attr('numb'):
            if not (self.pos.shape == temp.shape):
                raise ValueError("Shape if pos and numb are non-consistent")
            self.pos[:] = temp
        else:
            self.pos = temp
            self.numb = temp.shape[-1]

    def _init_link(self, link, linktype=None):
        link_local = link[:]
        if type(link_local) == np.ndarray:
            if linktype.shape[0] != 2 and link_local.shape[0] == 2:
                link_local = link_local.transpose()

        try:
            temp = []
            if linktype is not None:
                if len(linktype) != len(link_local):
                    raise ValueError("length of linktype if given should be same with length of link:link({}), linktype({})".
                                     format(len(link_local), len(linktype)))

            for one_link in link_local:
                if len(one_link) != 2:
                    raise TypeError("Wrong shape of links, need iterable of [x1,x2]")

                if (one_link[0] > self.numb - 1) or (one_link[1] > self.numb - 1):
                    raise ValueError("Link mismatch:{} - {} ".format(one_link[0], one_link[1]))

                if one_link[0] == one_link[1]:
                    raise RuntimeWarning("Ignored link:{} - {}".format(one_link[0], one_link[1]))

                if one_link[0] < one_link[1]:
                    temp.append(one_link)
                else:
                    temp.append([one_link[1], one_link[0]])

            temp_linktype = self.__default_attr(attr_name='linktype', size=len(temp))
            temp = np.array(temp)

            if linktype is not None:
                dtype = [('first', int), ('last',int), ('type','<U5')]
                ln = len(linktype)
                for_sort = [(temp[x][0], temp[x][1], linktype[x]) for x in range(ln)]
                for_sort = np.sort(np.array(for_sort, dtype=dtype), order='first')
                temp = np.array([[x[0], x[1]] for x in for_sort])

                temp_linktype = np.array([x[2] for x in for_sort])
            else:
                #temp = np.sort(np.array(temp), axis=0)
                temp = temp[temp[:,0].argsort()]
            self.link = temp#.transpose()

            self.linktype = temp_linktype

        except TypeError:
            raise TypeError("Wrong shape of links, need iterable")

    def _init_atoms(self, atoms):
        if self.numb is not None and len(atoms) != self.numb:
            raise TypeError("Length of atoms mismatch numb")
        try:
            temp = np.array(atoms)
            try:
                temp = temp.astype(np.dtype('<U2'))
            except ValueError:
                raise TypeError("Wrong type of atoms")
            if not (np.in1d(self.atoms, self.__class__.Atomskeys)).all():
                raise ValueError("Unknown atoms")
            self.atoms = temp
            self.numb = len(self.atoms)
        except TypeError:
            raise TypeError("Can't turn atoms to list of str")

    def _init_text(self, text):
        if self.numb != self.__default_attr('numb'):
            if len(text) != self.numb:
                raise TypeError("Length of text mismatch numb")

        try:
            self.text = np.array(text).astype(str)
            self.numb = len(self.text)
        except TypeError:
            raise TypeError("Can't turn text to list of str")

    def _init_linktype(self, linktype):
        if self.numb != self.__default_attr('numb'):
            if len(linktype) != self.numb:
                raise TypeError("Length of linktype mismatch numb")

        try:
            self.text = np.array(linktype).astype(str)
        except TypeError:
            raise TypeError("Can't turn text_adds to list of str")


    def __get_all_lengths(self):
        if self.numb is None:
            numb = 0
        else:
            numb = self.numb
        poslen = self.pos.shape[-1]
        atomslen = self.atoms.shape[-1]
        textlen = self.text.shape[-1]
        return [numb, poslen, atomslen, textlen]


    def _self_consistency(self):
        value = self.__get_all_lengths()
        name_value = ['numb', 'pos', 'atoms', 'text']
        base_value = []
        for name in name_value:
            base_value.append(0)

        nonbase, nonbase_name, base_name = _MolBase.__comparison(dpc(value), dpc(base_value), dpc(name_value))

        valueset = set(nonbase)
        if len(valueset)>1:
            msg = "Discrepancy: "
            for num, val in enumerate(nonbase):
                msg += "{}({}); ".format(nonbase_name[num], val)
            raise ValueError(msg)

        if len(valueset) == 0:
            return

        length = valueset.pop()

        for name in base_name:
            self.__default_attr(name, size=length, inplace=True)

        if not (np.in1d(self.atoms, self.__class__.Atomskeys)).all():
            raise ValueError("Unknown atoms")

if __name__ =="__main__":
    val = [1,1,0,0,5]
    base = [1,2,3,4,5]
    name = ['a', 'b', 'c', 'd', 'e']
    vals, vals_names, bases = comparison(val, base, name)

