import numpy as np
from numpy.linalg import norm

from mol2_pack import MolWrap


def merge_molwrap(mol1, mol2):
    """
    returns new mol with merged mol1 and mol2
    :param mol1: MolWrap first
    :param mol2: MolWrap second
    :return:
    """
    if not(type(mol1)==type(mol2)==MolWrap):
        msg = "Merge Mol should be uset to merge 2 Mol files.Given arg types:{}, {}".format(type(mol1), type(mol2))
        raise TypeError(msg)
    numb = mol1.numb + mol2.numb
    pos = np.concatenate((mol1.pos, mol2.pos), axis=1)
    atoms = np.concatenate((mol1.atoms, mol2.atoms))
    text = np.concatenate((mol1.text, mol2.text))
    link = np.concatenate((mol1.link, mol2.link+mol1.numb))#,axis=-1)
    linktype = np.concatenate((mol1.linktype, mol2.linktype))
    res = MolWrap(numb=numb, pos=pos, atoms=atoms, link=link, text=text, linktype=linktype)
    return res


def file_to_molwrap(filename):
    filemol2 = open(filename)
    lines = [line.strip('\n') for line in filemol2]
    numb = int(lines[3].split()[0])
    bond_num = int(lines[3].split()[1])
    constshift = 9
    pos = np.zeros((3, numb))
    atoms = np.zeros(numb, dtype=np.dtype((str, 2)))
    text = np.zeros(numb, dtype=np.dtype((str, 20)))
    atom_parts = 5
    pos_start = 2
    for num in range(numb):
        currline = lines[constshift+num].split(' ',atom_parts)
        atoms[num] = currline[1]

        try:
            pos[:, num] = np.array(currline[pos_start:atom_parts]).astype('float')
        except:
            raise ValueError("wrong type in file")
        text[num] = currline[atom_parts]

    bondshift = constshift+numb+1

    bonds = []
    bondstype = []
    for num in range(bond_num):
        currline = lines[bondshift+num].split()
        bonds.append([int(currline[1])-1, int(currline[2])-1])
        bondstype.append(currline[3])
    mol = MolWrap(numb=numb, pos=pos, atoms=atoms, text=text, link=bonds, linktype=bondstype)
    return mol


def search_Na(mol):
    ind_na = np.where(mol.atoms=="Na")
    if len(ind_na)>1:
        msg = "Several Na in molecule; SetNa_axis is supposed to be used only for single molecule. " \
              "mol wouldn't be changed"
        raise RuntimeWarning(msg)
        # noinspection PyUnreachableCode
        return

    if len(ind_na)<1:
        msg = "No Na in molecule; mol wouldn't be changed"
        raise RuntimeWarning(msg)
        return
    return ind_na[0]


def set_Na_axis(mol, axis):
    """
    Looks for Na in molecule and setdirection for Na atom.
    There should be only one Na in molecule
    :rtype: None
    :param mol: MolWrap to be directed
    :param:axis: direction for Na atom
    :return: None
    """
    ind_na = search_Na(mol)
    vinit = mol.pos[:, ind_na]
    vfinal = np.array(axis)
    rot_matrix = rotation_matrix(vinit, vfinal)
    mol.setPos(np.dot(rot_matrix, mol.pos))
    return


def set_Na_zero(mol):
    ind_na = search_Na(mol)
    mol.pos -= mol.pos[:,ind_na]
    return


def rotation_matrix(vinit, vfinal):
    """
    Rodrigues'_rotation_formula: Rotation Matrix R, (vfinal = R*vinit)
    :param vinit: initial vector
    :param vfinal: final vector
    :return: R
    """
    vi = np.copy(vinit).reshape(3)
    vf = np.copy(vfinal).reshape(3)
    vi /= norm(vi)
    vf /= norm(vf)
    nrot = np.cross(vi, vf)
    nrot /= norm(nrot)
    theta = np.arccos(np.vdot(vi, vf)/(norm(vf)*norm(vi)))
    K = np.array([[0., -nrot[2], nrot[1]],
                  [nrot[2], 0, -nrot[0]],
                  [-nrot[1], nrot[0], 0]
    ])

    R = np.eye(3) + np.sin(theta)*K + (1-np.cos(theta))*(np.dot(K, K))
    return R


def molwrap_to_file(mol, filename, name="mol2_pack_sketch"):
    filemol2 = open(filename,"w")
    f = filemol2
    numb = mol.numb
    bond_num = len(mol.link)
    pos = mol.pos
    atoms = mol.atoms
    text = mol.text
    link = mol.link
    linktype = mol.linktype
    f.write("#generated by mol2_pack\n")
    f.write("@<TRIPOS>MOLECULE\n")
    f.write(name+"\n")
    namestr = "{} {} 0 0\n".format(numb, bond_num)
    f.write(namestr)
    f.write("SMALL\n")
    f.write("USER_CHARGES\n")
    f.write("\n")
    f.write("\n")

    f.write("@<TRIPOS>ATOM\n")
    for n in range(numb):
        cpos = pos[:,n]
        currstring = "{} {} {:.6f} {:.6f} {:.6f} {}\n".format(n+1, atoms[n], cpos[0], cpos[1], cpos[2], text[n])
        f.write(currstring)

    f.write("@<TRIPOS>BOND\n")
    for n in range(bond_num):
        currbond = link[n]
        currstring = "{} {} {} {}\n".format(n+1, currbond[0]+1, currbond[1]+1,linktype[n])
        f.write(currstring)
    f.close()

if __name__=="__main__":
    filin = "../ex.mol2"
    mol = file_to_molwrap(filin)

    gen= "../data/"
    filout = gen+"mol_pack.mol2"
    molwrap_to_file(mol, filout)


    filout2 = gen+"mol_pack_mean.mol2"
    mol2 = mol.copy()
    mol2.meanPos()
    molwrap_to_file(mol2, filout2)

    filout3 = gen+"mol_pack_rotate.mol2"
    mol2 = mol.copy()
    R = rotation_matrix(np.array([1., 0., 0,]), np.array([0., 1., 0.]))
    mol2.rotate(R)
    molwrap_to_file(mol2, filout3)

    filout4 = gen+"mol_pack_double.mol2"
    mol2 = merge_molwrap(mol, mol2)
    molwrap_to_file(mol2, filout4)
